// SPDX-License-Identifier: MIT
pragma solidity ^0.8.22;

import "@openzeppelin/contracts-upgradeable/access/OwnableUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/proxy/utils/Initializable.sol";

contract MinterAble is Initializable, OwnableUpgradeable {
    mapping(address => bool) private minters;

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyMiner() {
        require(minters[msg.sender], "!miner");
        _;
    }

    function minter_initialize() public initializer {
        __Ownable_init(msg.sender); // Initialize with the owner
    }

    function addMinter(address _minter) public onlyOwner {
        minters[_minter] = true;
    }

    function removeMinter(address _minter) public onlyOwner {
        minters[_minter] = false;
    }
}



contract ArrayCheck is Initializable {

    uint256 internal BATCH_COUNT;
    function ac_initialize() public onlyInitializing {
        BATCH_COUNT = 256;
    }

    modifier checkArraySizeUint256(uint256[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySizeAddress(address[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySizeString(string[] memory x) {
        require(x.length > 0 && x.length <= BATCH_COUNT, "Too many elements in the array");
        _;
    }

    modifier checkArraySize1(uint256 len1) {
        require(len1 > 0 && len1 <= BATCH_COUNT, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize2(uint256 len1, uint256 len2) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize3(uint256 len1, uint256 len2, uint256 len3) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize4(uint256 len1, uint256 len2, uint256 len3, uint256 len4) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize5(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize6(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5, uint256 len6) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5 && len5 == len6, "The size of These Arrays must be equal");
        _;
    }

    modifier checkArraySize7(uint256 len1, uint256 len2, uint256 len3, uint256 len4, uint256 len5, uint256 len6, uint256 len7) {
        require(len1 > 0 && len1 <= BATCH_COUNT && len1 == len2 && len2 == len3 && len3 == len4 && len4 == len5 && len5 == len6 && len6 == len7, "The size of These Arrays must be equal");
        _;
    }
}

contract Admin is Initializable, OwnableUpgradeable, ArrayCheck {
    mapping(address => bool) private _admins;

    modifier onlyAdmin() {
        /*
        if (!_admins[msg.sender]) {
            console.log("onlyAdmin", address(this), msg.sender);
        }
        */
        require(_admins[msg.sender], "Only admin can call it");
        _;
    }

    function admin_initialize() internal onlyInitializing {
        __Ownable_init(msg.sender); // Initialize with the owner
        _admins[msg.sender] = true; // Set the initial owner as admin
        ac_initialize();
    }

    function addAdmins(address[] memory admins_) public onlyOwner checkArraySizeAddress(admins_) {
        for (uint256 i = 0; i < admins_.length; i++) {
            addAdmin(admins_[i]);
        }
    }

    function delAdmins(address[] memory admins_) public onlyOwner checkArraySizeAddress(admins_) {
        for (uint256 i = 0; i < admins_.length; i++) {
            delAdmin(admins_[i]);
        }
    }

    function addAdmin(address admin_) public onlyOwner {
        _admins[admin_] = true;
        // console.log("addAdmin", address(this), _admin);
    }

    function delAdmin(address admin_) public onlyOwner {
        _admins[admin_] = false;
        // console.log("delAdmin", address(this), _admin);
    }

    function isAdmin(address _addr) public view returns (bool) {
        return _admins[_addr];
    }
}

contract Program is Admin {
//    using Address for address;
    mapping(address => bool) private _programs;

    modifier onlyProgram() {
        require(_programs[msg.sender], "Only program can call it");
        _;
    }

    function program_initialize() internal onlyInitializing {
        admin_initialize();
    }

    function addPrograms(address[] memory programs_) public onlyOwner checkArraySizeAddress(programs_) {
        for (uint256 i = 0; i < programs_.length; i++) {
            addProgram(programs_[i]);
        }
    }

    function delPrograms(address[] memory programs_) public onlyOwner checkArraySizeAddress(programs_) {
        for (uint256 i = 0; i < programs_.length; i++) {
            delProgram(programs_[i]);
        }
    }

    function addProgram(address _program) public onlyOwner {
        //require(_program.isContract(), "Only contract can be a program");
        _programs[_program] = true;
    }

    function delProgram(address program_) public onlyOwner {
        _programs[program_] = false;
    }

    function isProgram(address program_) public view returns (bool) {
        return _programs[program_];
    }
}
